/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * </p>The class {@link AnnotationEntry} is responsible for providing those annotations of a given
 * model element that are relevant for the current view. It should be noted that annotation views
 * should never work directly on {@link IAnnotatedSpecification}s, but use {@link AnnotationEntry}s
 * as an interface to the model.</p>
 * 
 * <p>
 * {@link AnnotationEntry}
 * </p>
 * <ul>
 * <li>maintains a {@link #specificationsList} that keeps track of all the specifications attached
 * to a model element.</li>
 * <li>provides methods that delegate the access to the underlying annotations to the registered
 * {@link IAnnotatedSpecification}s.</li>
 * </ul>
 * 
 * @author eder, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 05CDA528EF2A70AD9E35B0ACCF2D21C9
 */
public final class AnnotationEntry {

	/**
	 * Model element for which the set of relevant {@link IAnnotatedSpecification}s is managed by
	 * this {@link AnnotationEntry}.
	 */
	private IModelElement modelElement;

	/** Hash map {@link IAnnotatedSpecification} -> {@link IAnnotationValueProvider}. */
	HashMap<Class<? extends IAnnotatedSpecification>, IAnnotationValueProvider<IAnnotatedSpecification>> providerSpecMapping =
			new HashMap<Class<? extends IAnnotatedSpecification>, IAnnotationValueProvider<IAnnotatedSpecification>>();

	/** List of specifications of {@link #modelElement}. */
	private List<IAnnotatedSpecification> specificationsList =
			new ArrayList<IAnnotatedSpecification>();

	/** Constructs a new {@link AnnotationEntry}. */
	public AnnotationEntry(IModelElement modelElement) {
		this.modelElement = modelElement;
	}

	/**
	 * Predicate if the given {@link AnnotationEntry} is empty, i.e., if it does not contain any
	 * {@link IAnnotatedSpecification}s.
	 */
	public boolean isEmpty() {
		return specificationsList.isEmpty();
	}

	/**
	 * Predicate that checks if the {@link IAnnotatedSpecification}'s operation that is wrapped by
	 * this class {@link AnnotationEntry} is enabled.
	 */
	private boolean isOperationEnabled(IAnnotatedSpecification specification) {
		// Specification may not be hooked to an IModelElement when an undo / redo is in progress
		return specification != null && specification.getSpecificationOf() != null;
	}

	/** Adds a new tuple {@link IAnnotationValueProvider} {@link IAnnotatedSpecification}. */
	public void addNewSpecification(IAnnotationValueProvider<IAnnotatedSpecification> provider,
			IAnnotatedSpecification spec) {
		providerSpecMapping.put(spec.getClass(), provider);
		specificationsList.add(spec);
	}

	/**
	 * Returns the name (= label to be used in annotation views) for a given annotation
	 * {@code clazz}.
	 */
	public String getSpecificationAnnotationName(Class<? extends IAnnotatedSpecification> clazz) {
		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			return getAnnotationValueProvider(clazz).getAnnotationName(specification);
		}

		return null;
	}

	/** Sets the name (= label to be used in annotation views) for a given annotation {@code clazz}. */
	public void setSpecificationAnnotationName(String name,
			Class<? extends IAnnotatedSpecification> clazz) throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			getAnnotationValueProvider(clazz).setAnnotationName(name, specification);
		}

		throw new Exception("Could not find an AnnotationValueProvider for " + clazz.toString());
	}

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			return getAnnotationValueProvider(clazz).canEdit(specification);
		}

		return false;
	}

	/** Predicate if the given annotation can be edited. TODO(VA) for the given instanceKey? */
	public boolean canEdit(Class<? extends IAnnotatedSpecification> clazz, String instanceKey) {
		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			return getAnnotationValueProvider(clazz).canEdit(specification, instanceKey);
		}

		return false;
	}

	/** Returns the annotation value. */
	public <V> V getSpecificationValue(Class<? extends IAnnotatedSpecification> clazz,
			String instanceKey) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			try {
				return getAnnotationValueProvider(clazz).getAnnotationValue(specification,
						instanceKey);
			} catch(Exception e) {
				// Ignore exception
			}
		}

		return null;
	}

	/**
	 * Returns the annotation value for a specific instance for an annotation with multi-value
	 * support.
	 */
	public <V> V getSpecificationValue(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			try {
				return getAnnotationValueProvider(clazz).getAnnotationValue(specification);
			} catch(Exception e) {
				// Ignore exception
			}
		}

		return null;
	}

	/** Sets a new value for the given specification {@code clazz}. */
	public <V> void setSpecificationValue(V value, Class<? extends IAnnotatedSpecification> clazz,
			String instanceKey) throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification, instanceKey);
		} else {
			// TODO(VA) The following exception and message are repeated many times. Would be worth
			// factorizing? e.g., using a method "couldNotFindProviderExn(Class clazz)"
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/**
	 * Sets a new value for the given specification {@code clazz} from a {@link String}
	 * representation.
	 */
	public void setSpecificationValue(String value, Class<? extends IAnnotatedSpecification> clazz,
			String instanceKey) throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification, instanceKey);
		} else {
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/** Sets a new value for the given specification {@code clazz}. */
	public <V> void setSpecificationValue(V value, Class<? extends IAnnotatedSpecification> clazz)
			throws Exception {
		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification);
		} else {
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/***
	 * Sets a new value for the given specification {@code clazz} from a {@link String}
	 * representation.
	 */
	public void setSpecificationValue(String value, Class<? extends IAnnotatedSpecification> clazz)
			throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification);
		} else {
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/**
	 * Returns the instance keys supported by the respective {@link IAnnotationValueProvider}.
	 * {@link IAnnotationValueProvider}s can either define a static {@link List}, or enable the
	 * dynamic creation of instance keys using {@link #allowsDynamicAnnotationInstances(Class)}. In
	 * the latter case, this method returns current list of dynamic annotation instances.
	 */
	public Collection<String> getInstanceKeys(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			return getAnnotationValueProvider(clazz).getInstanceKeys(specification);
		}

		return Collections.emptyList();
	}

	/**
	 * Indicates whether the given specification contains a value type allowing the storage of
	 * dynamically created instances of the annotation (like a list of strings). Static instances
	 * can be implemented by overriding
	 * {@link IAnnotationValueProvider#getInstanceKeys(IAnnotatedSpecification)} and returning the
	 * {@link List} of admissible instance keys.
	 */
	public boolean allowsDynamicAnnotationInstances(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(isOperationEnabled(specification)) {
			return getAnnotationValueProvider(clazz).allowsDynamicAnnotationInstances();
		}

		return false;
	}

	/** Returns {@link #modelElement}. */
	public IModelElement getModelElement() {
		return modelElement;
	}

	/** Returns {@link #specificationsList}. */
	public List<IAnnotatedSpecification> getSpecificationsList() {
		return specificationsList;
	}

	/** Determine the value provider for the given annotation class. */
	public IAnnotationValueProvider<IAnnotatedSpecification> getAnnotationValueProvider(
			Class<? extends IAnnotatedSpecification> clazz) {

		return providerSpecMapping.get(clazz);
	}

	/**
	 * Returns the {@link IAnnotatedSpecification} of a given {@code clazz} managed by this
	 * {@link AnnotationEntry}.
	 */
	@SuppressWarnings("unchecked")
	public <T extends IAnnotatedSpecification> T getSpecification(Class<T> clazz) {
		for(IAnnotatedSpecification s : specificationsList) {
			if(clazz.isInstance(s)) {
				return (T)s;
			}
		}

		return null;
	}
}
